unit UnitFrmRemoved;

{
    Purpose:
        User interface to copy a removed item to the clipboard

    Updates:
        fixed broken "Copy" button
    --------------
        Made form scalable

    NOTES: I'm using the same dirty trick to work-around Win9x control
    memory limitations as I used in FrmClipboardManager.
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, UnitClipQueue, ComCtrls, ExtCtrls, xpman;

type
  TFrmRemoved = class(TForm)
    bCopy: TButton;
    bClose: TButton;
    Label1: TLabel;
    mPreview: TMemo;
    bFlushItems: TButton;
    btnCopyAs: TButton;
    iPreview: TImage;
    lbList: TListBox;
    Label2: TLabel;
    rPreview: TRichEdit;
    FindDialog1: TFindDialog;
    btnFind: TButton;
    btnEdit: TButton;
    procedure FormShow(Sender: TObject);
    procedure lbListClick(Sender: TObject);
    procedure bCopyClick(Sender: TObject);
    procedure bCloseClick(Sender: TObject);
    procedure bFlushItemsClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormHide(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btnCopyAsClick(Sender: TObject);
    procedure lbListMeasureItem(Control: TWinControl; Index: Integer;
      var Height: Integer);
    procedure lbListDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);

    procedure FreeRemoved(ci : TClipItem);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure btnFindClick(Sender: TObject);
    procedure FindDialog1Find(Sender: TObject);
    procedure btnEditClick(Sender: TObject);
  private
    { Private declarations }
    //HistoryLoaded : boolean;
    ci : TClipItem;

  public
    { Public declarations }
  end;

var
  FrmRemoved: TFrmRemoved;

implementation

{$R *.dfm}

Uses StrUtils, Clipbrd, UnitFrmHotKey, UnitFrmConfig,
  UnitFrmClipboardManager, UnitPaste, Math, UnitMisc, UnitFrmEditItem;
const REMOVED_FILE = 'removed.txt';

procedure TFrmRemoved.FormShow(Sender: TObject);
var i : longint;
    s : string;
begin
    bCopy.Enabled := false;

    mPreview.Clear;

    lbList.Items.BeginUpdate;
    lbList.Items.Clear;
    for i := (RemovedQueue.GetQueueCount - 1) downto 0 do begin
        s := RemovedQueue.GetItemText(i);
        lbList.Items.Add(LeftStr(s,160));
    end;
    lbList.Items.EndUpdate;
end;

procedure TFrmRemoved.lbListClick(Sender: TObject);
var i : integer;
    s : string;
begin
    // reverse the index from 0 - (N-1) to (N-1) - 0
    bCopy.Enabled := true;
    mPreview.Visible := false;
    iPreview.Visible := false;
    rPreview.Visible := false;

    bCopy.Enabled := true;
    i := (lbList.Count - 1) - lbList.ItemIndex;
    mPreview.Text := Leftstr(
        RemovedQueue.GetItemText(i),
        10000);

    self.FreeRemoved(ci);
    ci := RemovedQueue.GetItemClip(i);

    btnCopyAs.Visible := false;
    if (ci <> nil) then begin
        if (ci.GetFormat <> 0) then begin
            btnCopyAs.Hint := ci.GetFormatName;
            btnCopyAs.Caption := 'Copy as ' + LeftStr(btnCopyAs.Hint, 10);

            btnCopyAs.Visible := true;
        end;

        btnEdit.Visible := mPreview.Text <> '';

        if (ci.GetFormat = CF_DIB) then begin
            mPreview.Visible := false;
            iPreview.Visible := true;
            ci.GetDIB(iPreview.Picture);
        end;
        if (ci.GetFormat = UnitMisc.GetCF_RICHTEXT) then begin
            mPreview.Visible := false;
            iPreview.Visible := false;
            rPreview.Visible := true;
            ci.GetRichText(s);
            rPreview.Text := s;
        end;
    end;

    if(not iPreview.Visible) and (not rPreview.Visible)then begin
        mPreview.Visible := true;
    end;
end;

procedure TFrmRemoved.bCopyClick(Sender: TObject);
var i : integer;
begin
    //Paste.SendText(PChar(lbList.Items.Objects[lbList.itemindex]));

    Paste.SetClipboardOnlyOnce;
    i := (lbList.Count - 1) - lbList.ItemIndex;

    if frmconfig.cbMoveFirst.checked then begin
        frmClipboardManager.DisablePasteProtectionOnce;
    end;

    Paste.SendText(RemovedQueue.GetItemText(i), nil);
    
    self.hide;
end;

procedure TFrmRemoved.bCloseClick(Sender: TObject);
begin
    lblist.Items.Clear;
    self.Close;
end;

procedure TFrmRemoved.bFlushItemsClick(Sender: TObject);
begin
    RemovedQueue.ClearQueue;
    lbList.Items.Clear;
    self.mPreview.Clear;
    self.iPreview.Visible := false;
end;


procedure TFrmRemoved.FormClose(Sender: TObject; var Action: TCloseAction);
begin
    lblist.Items.Clear;
    mPreview.Clear;
    
    {
    if FrmConfig.cbRemember.checked then begin
        frmClipboardManager.SaveHistory;
    end;
    }
end;

procedure TFrmRemoved.FormHide(Sender: TObject);
begin
    lbList.Items.Clear;
end;

procedure TFrmRemoved.FormCreate(Sender: TObject);
begin
    UnitMisc.AppendLog('FrmRemoved - creating');

    //ci := TClipItem.Create;
    //lblist.style

    // lbOwnerDrawVariable is buggy with smooth scrolling
    // this is a workaround so it smoothscrolls in the right direction but
    // allows for a larger height item
    lblist.IntegralHeight := false;
    lblist.ItemHeight := math.max(17, lbList.Canvas.TextHeight('ALKJDIOJ')+3);
end;

procedure TFrmRemoved.btnCopyAsClick(Sender: TObject);
begin
    Paste.SetClipboardOnlyOnce;
    Paste.SendText('', ci);
end;

procedure TFrmRemoved.lbListMeasureItem(Control: TWinControl;
  Index: Integer; var Height: Integer);
begin
  height := math.max(16, lbList.Canvas.TextHeight('ALKJDIOJ')+4);
end;

procedure TFrmRemoved.lbListDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);
var ci : TClipItem;
    th, h : integer;
    c : TColor;
    s, s2 : string;
    dt : TDateTime;
begin
    // height of rectangle & height of text on the listbox
    h := (rect.Bottom - rect.top) + 1;
    th := lbList.Canvas.TextHeight('ALKJDFIOJ');


    //
    // draw the icon leaving enough pixels above and below
    // for the selection outline
    //
    s := lbList.Items.Strings[index];
    ci := RemovedQueue.GetItemClip((lbList.Count - 1) - index);
    dt := RemovedQueue.GetItemDate((lbList.Count - 1) - index);
    if (ci <> nil) then begin
        DrawIconEX(
            lbList.Canvas.Handle, rect.Left+2, rect.Top+1,ci.CData.GetHICON,
            h-2, h-2,
            0,0,DI_NORMAL);

        DateTimeToString(s2, 'mm-dd-yy', dt);
        s := UnitMisc.GetCliptypeSymbol(ci.GetFormat) + ' ' + s2 + ':   ' + s;

        self.FreeRemoved(ci);
    end;

    // leave a space for the icon
    // center the text if possible
    if (h  <=  th) then begin
        lbList.Canvas.TextOut(
            rect.Left + h + 3, rect.Top, s
        );
    end else begin
        lbList.Canvas.TextOut(
            rect.Left + h + 3, floor((h-th)/2) + rect.Top, s
        );
    end;


    //
    // Draw/Clear the selection outline
    //
    c := lbList.Canvas.Brush.color;
    if (odSelected in state) then begin
        lbList.Canvas.Brush.Color := clHighlight;
        lbList.Canvas.FrameRect(rect);
    end else begin
        lbList.Canvas.Brush.color := clWindow;
        lbList.Canvas.FrameRect(rect);
    end;
    lbList.Canvas.Brush.color := c;

end;
procedure TFrmRemoved.FreeRemoved(ci: TClipItem);
var h : HICON;
begin
    //
    // Removed items are generated on the fly and resources must
    // be freed manually.
    //
    if (ci <> nil) then begin
        h := ci.CData.GetHICONAbsolute;
        UnitMisc.MyDestroyIcon(h);
        MyFree(ci);
    end;
end;

procedure TFrmRemoved.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    if (KEY = VK_ESCAPE) then begin
        self.bClose.Click;
    end;
end;

procedure TFrmRemoved.btnFindClick(Sender: TObject);
//var y : integer;
begin
    {
    y := mouse.CursorPos.Y;
    FindDialog1.Position := point(FindDIalog1.Position.x, y);
    }
    FindDialog1.Execute; 
end;

procedure TFrmRemoved.FindDialog1Find(Sender: TObject);
var i,j, a,b : integer;
    s : string;
    pc : PChar;
    op : TStringSearchOptions;
begin
    if lblist.Count = 0 then EXIT;

    // select the first item in the list by default
    // translate the index because RemovedCach is stored oldest first
    if lbList.itemindex = -1 then begin
        lblist.ItemIndex := 0;
    end;
    i := lbList.ItemIndex;
    i := (lbList.Count-1) - i;

    op := [soDown];
    if frMatchCase in FindDialog1.options then op := op + [soMatchCase];
    if frwholeword in FindDialog1.options then op := op + [soWholeWord];

    // in an oldest-first list, down and up searches are reversed
    if (frdown in FindDialog1.Options) then begin
        a := 0;
        b := i;
        for j := b-1 downto a do begin
            s := RemovedQueue.GetItemText(j);
            pc := SearchBuf(
                pchar(s), length(s),
                0,0, FindDialog1.FindText,
                op
            );
            if pc <> nil then begin
                lblist.ItemIndex := (lblist.count-1)-j;
                lblist.TopIndex := (lblist.count-1)-j;
                EXIT;
            end;
        end;
    end else begin
        a := i;
        b := lbList.Count - 1;

        for j := a+1 to b do begin
            s := RemovedQueue.GetItemText(j);
            pc := SearchBuf(
                pchar(s), length(s),
                0,0, FindDialog1.FindText,
                op
            );
            if pc <>nil then begin
                lblist.ItemIndex := (lblist.count-1)-j;
                lblist.TopIndex := (lblist.count-1)-j;
                EXIT;
            end;
        end;
    end;
end;


procedure TFrmRemoved.btnEditClick(Sender: TObject);
var s : string;
    i : integer;
begin
    if ci = nil then EXIT;
    if lblist.ItemIndex < 0 then EXIT;

    s := ci.GetAsText;
    if s = '' then begin
        s := RemovedQueue.GetItemText((lblist.count-1)-lbList.ItemIndex);
    end;
    FrmEditItem.SetText(s, self);

    //FrmEditItem.Position := poOwnerFormCenter;
    i := FrmEditItem.ShowModal;
    //FrmEditItem.Position := poDesigned;

    if i <> mrCancel then begin
        self.FormShow(self);
    end;
end;

end.


