unit UnitMenuItemTagdata;
{
Purpose:
    There are only so many dirty tricks you can do with ImageIndex, Tag, and Hint
    before you actually have to program it the correct way in the first place.

    This unit associates data with a menu item on the popup.
}

interface

uses Contnrs, UnitClipQueue, Windows;

    type TItemType = (
        IT_NONE,
        IT_LINE,
        IT_MENU,
        IT_OTHER,
        IT_PERMANENT,
        IT_PERMANENT_GROUP,
        IT_POPUP,
        IT_TEMP,
        IT_SWITCH,
        IT_SYS_SWITCH,
        IT_SYS_METHOD,
        IT_CURRENT,
        IT_SELECTED,
        IT_LAST,
        IT_PROGRAM,
        IT_SYSTEM
    );

    type TMenuItemTagData = class(TObject)
        caption : string;
        name : string;
        text : string;
        clip : TClipItem;
        prefix : string;
        itemtype : TItemType;
        itemindex : integer;
        icon : HICON;
        //isPicture : boolean;
        PermanentGroupID : integer;

        public
            constructor Create();
    end;


    type TTagDataList = class(TObject)
        private
            list : TObjectList;
        public
            constructor Create();
            procedure Flush;
            function NewTagData : integer;
            function Get(index : integer) : TMenuItemTagData;
            function ValidIndex(index : integer) : boolean;
    end;


    var TagDataList : TTagDataList;

implementation

uses SysUtils{for Exception};
{ TMenuItemTagData }

constructor TMenuItemTagData.Create;
begin
    caption := '';
    name := '';
    text := '';
    prefix := '';
    clip := nil;
    itemtype := IT_NONE;
    itemindex := -1;
    icon := 0;
    //isPicture := false;
    PermanentGroupID := -1;
end;


{ TTagDataList }

const MY_TAGDATA_OFFSET = 5000;

// now that time has passed, I've no clue why I used this offset
// I probably used lower values for the .Tag field for a separate meaning

constructor TTagDataList.Create;
begin
    list := TObjectList.Create;
    list.OwnsObjects := true;
end;

procedure TTagDataList.Flush;
begin
    list.Clear;
end;

function TTagDataList.Get(index: integer): TMenuItemTagData;
begin
    try
        result := TMenuItemTagData(  list.Items[index - MY_TAGDATA_OFFSET ] );
    except
        //result := TMenuItemTagData.Create();
        raise Exception.Create('TagData Index out of bounds');
        // TODO: real exception handling done here
        // this creates a memory leak
    end;
end;

function TTagDataList.NewTagData: integer;
begin
    result := list.Add( TMenuItemTagData.Create ) + MY_TAGDATA_OFFSET;
end;




function TTagDataList.ValidIndex(index: integer): boolean;
begin
    result := true;
    try
        self.Get(index)
    except
        result := false;
    end;
end;

initialization
    TagDataList := TTagDataList.Create;
end.

