unit UnitFrmDummyUnicodePopup;
{
    Purpose:
        Wrap all the logic needed to to create a low level API popup
        menu that has Unicode text and gather the data needed
        to show a Unicode Tooltip hint (see FrmDummyUnicodeTooltip

    Updates:

        Code completely re-written and simplified - FrmMainPopup's popup
        menu it traversed and duplicated instead of creating it from scratch
        
        -------------

        Show tooltip for "All Items..." hover
        ------------------------
        "All Items..." support
        ------------------------

        "Number of characters to show for each item" required a restart to
        work with the unicode popup

        ----------------------
        Let Permanent Items grow as large as they want
        "Beep" on accelerator key

}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StrUtils, INIFIles, Math, Menus,

  UnitMisc, UnitClipQueue, UnitFrmMainPopup, UnitTWideChar, 
  UnitOtherQueue, UnitKeyboardQuery, UnitFrmDummyUnicodeTooltip, Contnrs,
  StdCtrls,
  UnitFrmTooltip;

type
  TFrmDummyUnicodePopup = class(TForm)
    tim: TTimer;
    Label1: TLabel;
    tim2: TTimer;
    TimClose: TTimer;
    procedure timTimer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure tim2Timer(Sender: TObject);
    procedure TimCloseTimer(Sender: TObject);
  private
    { Private declarations }
    LastCursorPos : TPoint;
    LastIdlePos : TPoint;
    LastHint : string;
    LastIsMenuItem : boolean;
    LastClipItem : TClipItem;
    LastMenuItem : TMenuItem;
    LastSelectID : integer;
    LastMenuRect : Trect;
    LastHotkeyName : string;
    CurrentClipboard : TClipItem;

    KeyToItemID : THashedStringList;
    //VisibleCharCount : integer;
    //Tooltip : TTooltipWindow;
    Tooltip : TTooltipWindow;
    PasteTooltip : TTooltipWindow;
    TimeFiring : boolean;

    SubMenuItemList : THashedStringList;
    SubObjectList : TObjectList;

    mi : TMenuITem;
    m : TWMCommand;

    delayCommand : boolean;

    InClipTypeIcon : boolean;
    InApplicationIcon : boolean;

    RightClicked : boolean;
    TabCommand : boolean;

    pop : HMENU;

    function ItemIDToMenuPosition(itemID : cardinal) : integer;
    function ItemIDToMenuItem(itemID : cardinal) : TMenuItem;
    function MouseInCliptypeIcon : boolean;
    function MouseInApplicationIcon : boolean;
    procedure DisplayTootlip;
  public
    { Public declarations }

    procedure PopulatePopupUnicode(var M : TMenuItem; parent : HMENU);
    procedure ShowPopup(x,y : integer; inpopup : TPopupMenu);

    procedure WMDrawItem(var Msg: TWMDrawItem); message WM_DRAWITEM;
    procedure WMMeasureItem(var Msg: TWMMeasureItem); message WM_MEASUREITEM;
    procedure WMCommand(var Msg : TWMCommand); message WM_COMMAND;
    procedure WMMenuChar(var Msg : TWMMenuChar); message WM_MENUCHAR;
    procedure WMMenuSelect(var Msg : TWMMenuSelect); message WM_MENUSELECT;
    procedure WMEnterIdle(var Msg : TWMEnterIdle); message WM_ENTERIDLE;
  end;


    
var
  FrmDummyUnicodePopup: TFrmDummyUnicodePopup;
{////////////////////}
{//}implementation{//}
{////////////////////}
{$R *.dfm}


uses UnitFrmPermanentNew, UnitFrmConfig, UnitFrmSysTrayMenu, UnitPaste,
  Types, UnitMenuItemTagdata, UnitFrmEditItem, UnitPopupGenerate;
const //PU_LAST_TEXT = 'Las&t: ';
PU_LAST_TEXT = 'Last: ';
      //PU_CURRENT_TEXT = 'Cu&rrent: ';
PU_CURRENT_TEXT = 'Current: ';


CONST LEFT_COLUMN_WIDTH = 22;
CONST ICON_SPACER = LEFT_COLUMN_WIDTH + 4;

procedure TFrmDummyUnicodePopup.DisplayTootlip;
var ci :TClipItem;
    pos :    TPoint;
begin
    pos := mouse.CursorPos;

    if not Windows.PtInRect(self.LastMenuRect, mouse.CursorPos) then begin
        pos.X := self.LastMenuRect.left + ((self.LastMenuRect.right - self.lastmenurect.Left)div 2) ;
        pos.Y := self.LastMenuRect.bottom;
    end;

    pos.y := self.LastMenuRect.Bottom;
    inc(pos.x, 30);
    inc(pos.y, 5);
    UnitMisc.AppendLog('Showing Tooltip hint');

    InClipTypeIcon := false;
    InApplicationIcon := false;


    if self.MouseInCliptypeIcon then begin
        Tooltip.CloseTooltip;
        PasteToolTip.ShowTooltip(
            'Press TAB or Click here to paste as ' + self.LastClipItem.GetFormatName,
           pos, false );
        InClipTypeIcon := true;
    end else if self.MouseInApplicationIcon  then begin
        Tooltip.CloseTooltip;
        PasteTooltip.ShowTooltip('Click here to Edit', pos, false);
        InApplicationIcon := true;
    end else if (self.LastClipItem <> nil) then begin
        PasteTooltip.CloseTooltip;
        Tooltip.ShowTooltip(self.LastClipITem, pos,self.LastHotkeyName);
    end else begin
        PasteTooltip.CloseTooltip;

        if (FrmPermanent.IsComplexItem(self.LastHint)) then begin
            ci := FrmPermanent.GetComplexItem(self.LastHint);
            Tooltip.ShowTooltip(ci, pos, self.LastHotkeyName );
            MyFree(ci);
        end else begin
            Tooltip.ShowTooltip(self.LastHint, pos, true, self.LastHotkeyName );
        end;
    end;
end;

procedure TFrmDummyUnicodePopup.FormCreate(Sender: TObject);
begin
    UnitMisc.AppendLog('FrmDummyUnicodePopup - creating');

    KeyToItemID := THashedStringList.Create;
    CurrentClipboard := TClipItem.Create;

    LastClipItem := nil;
    
    Tooltip := TTooltipWindow.Create;;
    PasteTooltip := TTooltipWindow.Create;

    self.SubMenuItemList := THashedStringList.Create;
    SubObjectList := TObjectList.Create;
    SubObjectList.OwnsObjects := false;


    mi := TMenuItem.Create(self);
end;

procedure TFrmDummyUnicodePopup.FormDestroy(Sender: TObject);
begin

    SubMenuItemList.Free;
    SubObjectList.Free;


    mi.Free;
    tim.Enabled := false;
    tim2.Enabled := false;
end;


procedure TFrmDummyUnicodePopup.ShowPopup(x, y: integer;  inpopup: TPopupMenu);
var //p : HMENU;
    m : TMenuItem;
begin
    Windows.SetLastError(ERROR_SUCCESS);
    //self.VisibleCharCount := FrmConfig.udCharacters.Position;
    self.RightClicked := false;
    self.LastMenuRect := rect(0,0,0,0);

    self.LastIsMenuItem := false;
    FrmMainPopup.SetItemMenuOverride(false);
    InApplicationIcon := false;
    
    UnitMisc.AppendLog('translate');
    pop := Windows.CreatePopupMenu;
    m := inpopup.items;

    UnitMisc.AppendLog('current');
    self.CurrentClipboard.GetClipboardItem(0);
    

    self.PopulatePopupUnicode(m, pop);

    UnitMisc.AppendLog('track');
    // show it
    Windows.TrackPopupMenu(pop,
        TPM_LEFTBUTTON or TPM_RIGHTBUTTON or TPM_LEFTALIGN,
        x, y, 0,
        self.Handle, nil);

    UnitMisc.AppendLog('draw');
    Windows.DrawMenuBar(pop);

    UnitMisc.AppendLog('cleanup');

    // clean up
    while Windows.GetMenuItemCount(pop) > 0 do begin
        Windows.DeleteMenu(pop,0, MF_BYPOSITION);
    end;


    Tooltip.CloseTooltip;
    PasteTooltip.CloseTooltip;
    tim.Enabled := false;
    timclose.Enabled := false;
end;

// Call Order
//--------------
// Populate -> Measure Item -> Draw Item
// Some user event trigers
// (MenuSelect, EnterIdle, Command)

procedure TFrmDummyUnicodePopup.PopulatePopupUnicode(var M : TMenuItem; parent : HMENU);
    procedure TraverseMenuItems(var m : TMenuItem; parent : HMENU; level : integer = 0) ;
    var i : integer;
        menu : HMENU;
        m2 : TMenuItem;
        s : string;
        ni, ni2 : TNotifyEvent;
    begin
        if (m.count = 0) then begin
            if (m.IsLine) then begin


                i := SubObjectList.Add(m);
                Windows.AppendMenu(parent,MF_OWNERDRAW or MF_SEPARATOR or MF_BYCOMMAND, i, nil);
            end else begin
                i := SubObjectList.Add(m);
                ni := m.OnClick;
                ni2 := frmSysTrayMenu.MethodMenuItemClickEvent;
                if (m.Checked) then begin
                    // checked items are not custom drawn
                    // todo: implement this?
                    s := m.Caption;
                    Windows.AppendMenu(parent,
                        MF_ENABLED or MF_BYCOMMAND or MF_CHECKED,
                        i, PChar(s));
                end else begin
                    Windows.AppendMenu(parent, MF_OWNERDRAW, i, nil);
                end;
            end;
        end else begin
            menu := 0;
            if (level <> 0) then begin
                menu := Windows.CreatePopupMenu;
                UnitMisc.MySetTheme(menu);
            end;
            for i := 0 to (m.Count - 1) do begin
                m2 := m.Items[i];
                if (level <> 0) then begin
                    TraverseMenuItems(m2, menu, level + 1);
                end else begin
                    TraverseMenuItems(m2, parent, level + 1);
                end;
            end;
            if (level <> 0) then begin
                // the HMENU value is as the ID for drawing
                // instead of any user submitted number for submenu items (MF_POPUP)
                i := SubObjectList.Add(m);
                SubMenuItemList.Values[IntToStr(menu)] := IntToStr(i);
                Windows.AppendMenu(parent, MF_OWNERDRAW or MF_POPUP, menu, PChar(-1));
            end;
        end;
    end;
begin
    SubObjectList.Clear;
    SubMenuItemList.clear;
    TraverseMenuItems(m, parent, 0);

   
end;


{
//
// These 2 messages are just do draw the damn popu
//
}

procedure TFrmDummyUnicodePopup.WMMeasureItem(var Msg: TWMMeasureItem);
var id : Cardinal;
    s : string;
    sz : TSize;

    procedure GetTextSizeOf( s : string );
    begin
        Windows.GetTextExtentPoint32(self.Canvas.Handle, PChar(s),
        min(length(s), 200),
        sz);
        msg.MeasureItemStruct^.itemWidth := max(sz.cx + loword(GetMenuCheckMarkDimensions),20);
        msg.MeasureItemStruct^.itemHeight := max(sz.cy+1, 6);
    end;
    procedure GetTextSizeOfNoLimit( s : string );
    begin
        Windows.GetTextExtentPoint32(self.Canvas.Handle, PChar(s),
        length(s),
        sz);
        msg.MeasureItemStruct^.itemWidth := max(sz.cx + loword(GetMenuCheckMarkDimensions),20);
        msg.MeasureItemStruct^.itemHeight := max(sz.cy+1, 6);
    end;

var
    mi2 : TMenuItem;
    mitd : TMenuItemTagData;
begin
    //UnitMisc.AppendLog('wmmeasure');

    Windows.SetLastError(ERROR_SUCCESS);
    id := msg.MeasureItemStruct^.itemID;

    //
    // Basically, use the longest text item to gauge the width of the popup
    // At a minimum the height is the 2 pixels above the size of an icon
    //


    mi2 := self.ItemIDToMenuItem(id);
    s := '';
    if mi2 <> nil then begin
        s := mi2.Caption;
    end;



    msg.MeasureItemStruct^.itemWidth := 100;
    if (s <> '') then begin
        GetTextSizeOf(s);
    end else begin
        s := self.CurrentClipboard.GetAsText;
        GetTextSizeOf(PU_LAST_TEXT + s);
    end;

    try
        if (mi2 <> nil) and (mi2.tag <> 0) then begin

            mitd := TagDataList.Get(mi2.tag);
            case mitd.itemtype of
            IT_POPUP:
                begin
                    inc(msg.MeasureItemStruct^.itemWidth, loword(GetMenuCheckMarkDimensions));
                    msg.MeasureItemStruct^.itemHeight := max(msg.MeasureItemStruct^.itemHeight, 16);
                end;
            IT_MENU,  IT_TEMP,IT_OTHER, IT_CURRENT, IT_SELECTED, IT_LAST, IT_PROGRAM, IT_SYSTEM:
                begin
                    // account for ICON height
                    //inc(msg.MeasureItemStruct^.itemHeight, 4);
                    msg.MeasureItemStruct^.itemHeight := max(msg.MeasureItemStruct^.itemHeight, 16 + 2);
                    // small fonts can cause icon drawing problems
                end;
            IT_PERMANENT:
                    msg.MeasureItemStruct^.itemHeight := max(msg.MeasureItemStruct^.itemHeight, 16);

            end;
            if (mitd.icon <> 0) or (mi2.Bitmap.Height <> 0) then begin
                msg.MeasureItemStruct^.itemHeight := max(msg.MeasureItemStruct^.itemHeight, 16);
            end;
        end;
    except
    end;

    // ensure at least 100 pixels, but limit to Width Max setting
    msg.MeasureItemStruct^.itemWidth := max(msg.MeasureItemStruct^.itemWidth + LEFT_COLUMN_WIDTH, 100);
    msg.MeasureItemStruct^.itemWidth := min(frmconfig.UDWidth.Position, msg.MeasureItemStruct^.itemWidth);
end;


procedure TFrmDummyUnicodePopup.WMDrawItem(var Msg: TWMDrawItem);
var
    tc, bc : cardinal;
    dc : integer;
    TruncateCaption : boolean;
    selected : boolean;




    //
    // Since it's owner drawns, I have to know what &x character related
    // to what ItemID
    //
    procedure UpdateAccessKeys(s : string);
    var amp : integer;
    begin
        amp := pos('&', s);
        if (amp <> 0) then begin
            inc(amp);
            KeyToItemID.Values[lowercase(s[amp])] := IntToStr(msg.DrawItemStruct^.itemID);
        end;
    end;

    function GetTextOptions(s : string) : integer;
    const MY_TEXT_OPTIONS = DT_VCENTER or DT_SINGLELINE or
        DT_MODIFYSTRING or DT_END_ELLIPSIS;
    begin
        result := DT_VCENTER or DT_SINGLELINE;

        if TruncateCaption then begin
            result := result or DT_MODIFYSTRING {or DT_END_ELLIPSIS};
            if (pos('://',s) in [1..7])  then result := result or DT_PATH_ELLIPSIS and not DT_END_ELLIPSIS;
            if (pos(':\',s) in [1..7]) then result := result or DT_PATH_ELLIPSIS and not DT_END_ELLIPSIS;
        end;
    end;

    procedure DrawBitmap(hdc : HDC; bm : TBitmap; top, left, width, height : integer);
    var ico : TIcon;
        il : TImageList;
    begin
        if (bm <> nil) and ((bm.Width and bm.Height) <> 0) then begin
            ico := TIcon.Create;
            il := TImageList.CreateSize(bm.Width, bm.Height);
            il.AddMasked(bm, bm.TransparentColor);
            il.GetIcon(0, ico);
            il.Free;
            if (ico.Handle <> 0) then begin
                DrawIconEx(hdc,
                        left,
                        top,
                        ico.Handle,
                        width,height, 0, 0, DI_NORMAL);
            end;
            ico.Free;
        end;
    end;

    procedure DrawTextAndBitmap(bm : TBitmap; s : string);
    var r : TRect;
        width, height : integer;
    begin
        with msg.DrawItemStruct^ do begin
            UpdateAccessKeys(s);

            r.Top := rcItem.Top;
            r.Bottom := rcItem.Bottom;
            r.Left := rcItem.Left + ICON_SPACER;
            r.Right := rcItem.Right; //- (ICON_SPACER - 4);

            DrawText(hdc, PChar(s), length(s),
                r, GetTextOptions(s) and not DT_MODIFYSTRING );

            if bm = nil then EXIT;

            width := bm.width;
            height := bm.height;
            DrawBitmap(hdc, bm, rcItem.Top, rcItem.Left, Width , height);
        end;
    end;

    procedure DropShadowHorizontal(hdc : HDC; percent : double; y, x1, x2 : integer; OverrideColor : cardinal = 0 );
    var i : integer;
        c : cardinal;
        r, b, g : byte;

    begin
        for i := x1 to x2 do begin
            c := GetPixel(HDC,i, y);
            if OverrideColor <> 0 then c := OverrideColor;

            r := (c and $FF0000) shr 16;
            g := (c and   $FF00) shr 8;
            b := (c and     $FF);

            r := min(255, trunc(r * percent));
            g := min(255, trunc(g * percent));
            b := min(255, trunc(b * percent));

            c := RGB(r,g,b);
            SetPixel(HDC, i,y, c);
        end;
    end;
    procedure DropShadowVertical(hdc : HDC; percent : double; x, y1, y2 : integer; OverrideColor : cardinal = 0 );
    var i  : integer;
        c : cardinal;
        r, b, g : byte;
    begin
        for i := y1 to y2 do begin
            c := GetPixel(HDC, x,i);
            if OverrideColor <> 0 then c := OverrideColor;

            r := GetRValue(c);
            g := GetGValue(c);
            b := GetBValue(c);

            r := min(255, trunc(r * percent));
            g := min(255, trunc(g * percent));
            b := min(255, trunc(b * percent));

            c := RGB(r,g,b);
            SetPixel(HDC, x,i, c);
        end;
    end;

    procedure DropShadowBox(hdc : HDC; percent : double; x1,x2, y1,y2 : integer);
    var i,j : integer;
        c : cardinal;
        r,g,b : byte;
    begin
        for i := x2 downto x1 do begin
            for j := y2 downto y1 do begin
                c := GetPixel(HDC, i,j);
                if (c = 0) and (GetPixel(HDC, i+1, j+1) = $FFFFFF) then begin
                    //r := (c and $FF0000) shr 16;
                    //g := (c and   $FF00) shr 8;
                    //b := (c and     $FF);

                    r := min(255, trunc(255 * percent));
                    g := min(255, trunc(255 * percent));
                    b := min(255, trunc(255 * percent));

                    c := RGB(r,g,b);

                    SetPixel(HDC, i+1,j+1, c);
                end;

            end;
        end;

    end;

    procedure ClipTypeMark(hdc : HDC; rcItem : trect; cliptypeicon : tbitmap; oppacity : double = 1.0);
    var i : integer;
        top, left, width, height : integer;
        c : cardinal;
    begin
        if cliptypeicon = nil then EXIT;

        left :=  rcItem.Right   - cliptypeicon.Width - 3;
        top := rcItem.top + 1;
        width := cliptypeicon.Width;
        height := cliptypeicon.Height;

        cliptypeicon.TransparentColor := $AAAAAA; // unused color
        DrawBitmap(hdc, cliptypeicon, top, left, width, height);

        c := 0;
        DropShadowVertical(hdc, 0.9, left, top, top+height,c);
        DropShadowHorizontal(hdc,1.3,top, left,left+width,c);

        DropShadowVertical(hdc, 0.8, left+width-1, top+2, top+height,c);
        DropShadowHorizontal(hdc,0.8,top+height-1,left+2,left+width,c);
        if selected then begin
            c := ColorToRGB(clHighLightText);
        end;
        DropShadowVertical(hdc, 0.8, left+width, top+2, top+height,c);
        DropShadowHorizontal(hdc,0.8,top+height,left+2,left+width,c);


        if (oppacity < 1.0) then begin
            for i := top to top+height do begin
                DropShadowHorizontal(hdc, oppacity, i, left, left + width);
            end;
        end;
    end;

    procedure DrawUnicodeAndIcons(programicon : HICON; cliptypeicon : TBitmap; ci : TClipItem; prefix : string);
    var wc : TWideChar;
        r : TRect;
        sz : TSize;
        len : integer;
    begin
        UpdateAccessKeys(prefix);

        wc := TWideChar.Create;
        wc.Append(prefix);
        wc.Append(ci.GetHandle, ci.GetDataSize);

        //wc.Replace(WideChar($9), WideChar($3)); // replace these whitespace
        //wc.Replace(WideChar(#13),WideChar($3)); // characters with "squares"
        //wc.Replace(WideChar(#10),WideChar($3)); //

        {if (ci.GetDataSize / 2) > VisibleCharCount then begin
            wc.LeftStr(self.VisibleCharCount);
            wc.Append(' ...');
        end;
        }

        len := min(100, wc.StrLength -1);

        Windows.GetTextExtentPoint32W(self.Canvas.Handle, wc.Memory,
            wc.StrLength,sz);


        with msg.DrawItemStruct^ do begin
            r.Top := rcItem.Top + 1;
            r.Bottom := r.top + sz.cy;
            r.Left := rcItem.Left + ICON_SPACER;
            r.Right := rcItem.Right;
            dec(r.Right, ICON_SPACER);

            // trim the rectangle so the next line of text won't be
            // partially visible

            DrawTextW(hdc, PWideChar(wc.Memory), len,
                r, GetTextOptions(ci.GetAsText) );

            wc.Clear;
            MyFree(wc);

            if (programicon <> 0) then begin
                DrawIconEx(hdc,
                    rcItem.left, rcItem.top,
                    programicon,
                    16,16, 0, 0, DI_NORMAL);
            end;
        end;
    end;


    procedure DrawTextAndIcons(h : HICON; cliptypeicon : TBitmap; s : string); overload;
    var r : TRect;
    begin
        UpdateAccessKeys(s);

        with msg.DrawItemStruct^ do begin
        r.Top := rcItem.Top;
        r.Bottom := rcItem.Bottom + 1;
        r.Left := rcItem.Left + ICON_SPACER;
        r.Right := rcItem.Right - ICON_SPACER;

        DrawText(hdc, PChar(s), length(s),
            r, GetTextOptions(s));

        if (h <> 0) then begin
            DrawIconEx(hdc,
                    rcItem.left, rcItem.top,
                    h, 16,16,
                    0, 0, DI_NORMAL);
        end;


        ClipTypeMark(hdc, rcItem, cliptypeicon);
        end;
    end;

    procedure DrawTextAndIcon(h : HICON; s : string); overload;
    var r : TRect;
    begin
        UpdateAccessKeys(s);

        with msg.DrawItemStruct^ do begin
        r.Top := rcItem.Top;
        r.Bottom := rcItem.Bottom;
        r.Left := rcItem.Left + ICON_SPACER;
        r.Right := rcItem.Right;

        DrawText(hdc, PChar(s), length(s),
            r, GetTextOptions(s) );

        if (h <> 0) then begin
            DrawIconEx(hdc,
                    rcItem.left, rcItem.top,
                    h, 16,16,
                    0, 0, DI_NORMAL);
        end;
        end;
    end;

    function ClipTypeBitmap(ci : TClipItem) : TBitmap;
        function bm(img : TImage) : TBitmap;
        begin
            result := img.Picture.Bitmap;
        end;
    begin
        result := bm(FrmMainPopup.imgUnkown);
        case ci.GetFormat of
        CF_TEXT : result := bm(FrmMainPopup.imgPlain);
        CF_UNICODETEXT : result := bm(FrmMainPopup.ImgUnicode);
        CF_HDROP : result := bm(FrmMainPopup.imgFile);
        CF_DIB : result := bm(FrmMainPopup.imgPic);
        end;

        if ci.GetFormat = UnitMisc.GetCF_HTML then begin
            result := bm(FrmMainPopup.imgHTML);
        end else if ci.GetFormat = UnitMisc.GetCF_RICHTEXT then begin
            result := bm(FrmMainPopup.imgRichText);
        end;
    end;


    procedure DrawMenuItem(prefix : string; ci : TClipItem; icon : HICON = 0; instead : string = ''; noiconbutton : boolean = true); overload;
    var h : HICON;
        r, rPrefix, button : TRect;
        s : string;
        wc : TWideChar;
        sz : tagSIZE;
        oppacity : double;
        len : integer;
        br : HBRUSH;
    begin
        with msg.DrawItemStruct^ do begin
            h := 0;
            if icon <> 0 then begin
                h := icon;
            end else if (ci <> nil) then begin
                h := ci.CData.GetHICON;
            end;

            if (h <> 0)  then begin
                if selected then begin

                    if (not noiconbutton) then begin
                        br := Windows.CreateSolidBrush(ColorToRGB(clBtnFace));
                        button.top := rcitem.top;
                        button.bottom := rcitem.Bottom;
                        button.Left := 0;
                        button.Right := LEFT_COLUMN_WIDTH;
                        FillRect(hdc, button, br);
                        Windows.DeleteObject(br);
                    end;
                    DrawIconEx(hdc,
                        rcItem.left+2, rcItem.top+1,
                        h, 16,16,
                        0, 0, DI_NORMAL
                    );
                    if (not noiconbutton) then begin
                        button.right := 18+3;
                        Windows.DrawEdge(hdc,button,BDR_RAISEDinner, BF_TOP);
                        Windows.DrawEdge(hdc,button,BDR_RAISEDouter, BF_BOTTOM);
                        Windows.DrawEdge(hdc,button,BDR_RAISEDinner, BF_LEFT);
                        Windows.DrawEdge(hdc,button,BDR_RAISEDouter, BF_RIGHT);
                    end;
                end else begin
                    DrawIconEx(hdc,
                        rcItem.left+2, rcItem.top+1,
                        h, 16,16,
                        0, 0, DI_NORMAL
                    );

                end;
            end;

            rprefix := rect(0,0,0,0);
            if prefix <> '' then begin
                UpdateAccessKeys(prefix);

                s := prefix + '    ';
                DrawText(HDC, pchar(s), length(s), rPrefix, GetTextOptions(prefix) or DT_CALCRECT);
                s := trim(s) + ' ';
                DrawText(HDC, pchar(s), length(s), rPrefix, GetTextOptions(prefix) or DT_CALCRECT);

                rPrefix.Left := rcItem.left + ICON_SPACER + rPrefix.left ;
                rprefix.right := rcItem.Left + rPrefix.Right + ICON_SPACER;
                rPrefix.Top := rcItem.Top;
                rPrefix.Bottom := rcItem.Bottom;

                DrawText(hdc, PChar(s), length(s), rPrefix, GetTextOptions(prefix));
            end;
            r.top := rcItem.Top;
            r.bottom := rcItem.Bottom;
            r.Right := rcItem.Right - ICON_SPACER;
            r.left := rcItem.Left + ICON_SPACER + (rPrefix.right - rPrefix.left);
            if (instead = '')  and (ci <> nil) and (ci.GetFormat = CF_UNICODETEXT) then begin

                wc := TWideChar.Create;

                wc.Append(ci.GetHandle, ci.getdatasize);
                wc.Append(#0);
                len := min(100, wc.StrLength -2);
                sz.cx := 0;
                sz.cy := 0;
                DrawTextW(hdc, PWideChar(wc.Memory), len,
                    r, GetTextOptions(ci.GetAsText) );

                myfree(wc);
            end else begin
                s := '';
                if (ci = nil) then begin
                    if instead <> '' then
                        s := copy(instead,1,length(instead));
                end else begin
                    s := copy(ci.GetAsText,1,length(ci.GetAsText));
                end;
                // very strange - looks as if the compiler gives the address
                // of the object's string if "s := ci.GetAsText" - and the
                // data is altered
                if s <> '' then begin
                    len := min(100, length(s));
                    DrawText(hdc, PChar(s + ' '), len,
                        r, GetTextOptions(s) );
                end;

            end;
            oppacity := 1.0;

            if (ci <> nil) then begin
                if (ci.GetFormat in [CF_TEXT, CF_UNICODETEXT]) then begin
                    //oppacity := 0.6;
                end else begin
                    ClipTypeMark(hdc, rcItem, ClipTypeBitmap(ci), oppacity);
                end;
            end;
        end;

    end;



var HB, hb2, hb3, hb4, hb5 : HBRUSH;
    mi : TMenuItem;
    h : HGDIOBJ;
    oldf : HFONT;
    id, flags : integer;

    mitd : TMenuItemTagData;
    hilight, leftcolum, line, hotlight : TRect;

    vert : array[0..1] of TRIVERTEX;
    gRect   : GRADIENT_RECT;

begin
    Windows.SetLastError(ERROR_SUCCESS);
    //UnitMisc.AppendLog('wmdraw');



    with msg.DrawItemStruct^ do begin
    dc := Windows.SaveDC(hDC);
    if (dc = 0) then begin
        UnitMisc.AppendLog('WMDrawItem: SaveDC failed - ', true);
        EXIT;
    end;
    oldf := Windows.SelectObject(hdc, self.Font.Handle);

    id := msg.drawitemstruct.itemID;
    flags := Windows.GetMenuState(
        msg.drawitemstruct.hwndItem,
        self.ItemIDToMenuPosition(id),
        MF_BYPOSITION
    );

    //
    // Fill the entire background with a hilight or no higlight
    //

    selected := not ((itemState and ODS_SELECTED) = 0);
    if  not selected then begin
        tc := Windows.SetTextColor(hDC, ColorToRGB(font.Color));
        bc := Windows.SetBkColor(hDC, ColorToRGB(clMenu));
        hb := Windows.CreateSolidBrush(ColorToRGB(clMenu));
        hb2 := Windows.CreateSolidBrush(ColorToRGB(clHighLight));
    end else begin
        //tc := Windows.SetTextColor(hdc, ColorToRGB(clHighLightText));

        tc := Windows.SetTextColor(hdc, ColorToRGB(font.color));

        bc := Windows.SetBkColor(hdc, ColorToRGB(clHighlight));

        hb := Windows.CreateSolidBrush(ColorToRGB(clHighlight));

        hb2 := Windows.CreateSolidBrush(ColorToRGB(clMenu));
    end;
    hb5 := Windows.CreateSolidBrush(ColorToRGB(clHotLight));
    hb3 := Windows.CreateSolidBrush(ColorToRGB(cl3DLight));
    hb4 := Windows.CreateSolidBrush(ColorToRGB(clWindowText));

    // Does NOT work, only ODS_SELECTED ever seems to get sent
    if ((flags and (ODS_GRAYED or ODS_DISABLED {or ODS_INACTIVE})) <> 0) then begin
        tc := Windows.SetTextColor(hdc, ColorToRGB(clGrayText));
    end;
    if (hb <> 0) then begin

        if selected then begin
            SetBkMode(hdc, TRANSPARENT);
            Windows.FillRect(hdc,rcitem, hb5);
            hilight.Top := rcitem.Top+1;
            hilight.bottom := rcitem.bottom-1;
            hilight.Left := rcitem.Left + LEFT_COLUMN_WIDTH+2;
            hilight.Right := rcitem.Right-2;
            //Windows.FillRect(hdc, hilight, hb);


              vert [0] .x      := hilight.left;
              vert [0] .y      := hilight.top;
              vert [0] .Red    := GetRValue(ColorToRGB(clWindow)) shl 8;
              vert [0] .Green  := GetGValue(ColorToRGB(clWindow)) shl 8;
              vert [0] .Blue   := GetBValue(ColorToRGB(clWindow)) shl 8;
              vert [0] .Alpha  := $00;

              vert [1] .x      := hilight.Right;
              vert [1] .y      := hilight.bottom;
              vert [1] .Red    := GetRValue(ColorToRGB(clHighlight)) shl 8;
              vert [1] .Green  := GetGValue(ColorToRGB(clHighlight)) shl 8;
              vert [1] .Blue   := GetBValue(ColorToRGB(clHighlight)) shl 8;
              vert [1] .Alpha  := $00;

              gRect.UpperLeft  := 0;
              gRect.LowerRight := 1;
              GradientFill(hdc, @vert,2,@gRect,1,GRADIENT_FILL_RECT_V);

{
            hilight := rcItem;
            hilight.Right := LEFT_COLUMN_WIDTH;
            Windows.FillRect(hdc,hilight,hb3);

            hotlight.top := hilight.top+1;
            hotlight.bottom := hilight.bottom-1;
            hotlight.left := hilight.left+1;
            hotlight.right := hilight.right-1;
            Windows.FillRect(hdc, hotlight, hb);}

        end else begin
            Windows.FillRect(hdc, rcItem, hb);
        end;
        leftcolum.Right := LEFT_COLUMN_WIDTH;
        leftcolum.Left := 0;
        leftcolum.Top := rcitem.Top;
        leftcolum.Bottom := rcitem.Bottom;

        Windows.FillRect(hdc, leftcolum, hb3);

    end;

    try
        mi := self.ItemIDToMenuItem(itemID);
        TruncateCaption := false;
        try
            if mi.Tag <> 0 then begin

            mitd := TagDataList.Get(mi.tag);
            if mi <> nil then begin

                case mitd.itemtype  of
                IT_CURRENT:
                    begin
                        DrawMenuItem(mitd.prefix + PU_CURRENT_TEXT,self.CurrentClipboard, 0);
                    end;
                IT_SELECTED:
                    begin
                        DrawMenuItem(mitd.prefix, nil, 0, mitd.text);
                    end;
                IT_LAST:
                    begin
                        DrawMenuItem(mitd.prefix + PU_LAST_TEXT,nil, mitd.icon, mitd.text);
                    end;
                IT_MENU, IT_TEMP, IT_OTHER:
                    begin
                        DrawMenuItem(mitd.prefix, mitd.clip, mitd.icon,'', (mitd.itemtype = IT_TEMP));
                        TruncateCaption := true;
                    end;
                IT_PROGRAM, IT_SYSTEM:
                    begin
                        DrawMenuItem('',nil,mitd.icon,mi.caption);
                    end;
                IT_LINE:
                    begin
                        // todo: implement separater line

                        line.Top := rcitem.Top + trunc((rcitem.bottom - rcitem.top)/2);
                        line.Bottom := line.top +1;
                        line.Right := rcitem.Right-4;
                        line.Left := LEFT_COLUMN_WIDTH + 1;
                        // Vista-Like version
                        Windows.DrawEdge(hdc,line,BDR_SUNKENOUTER Or BDR_SUNKENINNER,BF_BOTTOM);
                        // darker XP-Like version
                        //Windows.DrawEdge(hdc,line,BDR_SUNKENOUTER or BDR_RAISEDINNER,BF_BOTTOM);
                    end;
                else
                    begin
                        DrawTextAndBitmap(mi.Bitmap, mitd.prefix + mi.caption);
                    end;                    
                end;

            end;
            end;
        except
            DrawTextAndBitmap(mi.Bitmap, mi.caption);
        end;

        Windows.DrawEdge(hdc,leftcolum,BDR_RAISEDINNER,BF_RIGHT);
        inc(leftcolum.Right);
        Windows.DrawEdge(hdc,leftcolum,BDR_SUNKENOUTER ,BF_RIGHT);

        rcitem.Left := 0;
        Windows.DrawEdge(hdc,rcitem,BDR_SUNKENINNER, BF_RIGHT);
        Windows.DrawEdge(hdc,rcitem,BDR_SUNKENOUTER, BF_RIGHT);

        //Windows.DrawEdge(hdc,rcitem,BDR_RAISEDINNER, BF_LEFT);

    finally
        h := SelectObject(hdc, oldf);
        {DeleteObject(h);
        }
        Windows.SetTextColor(hdc, tc);
        Windows.SetBkColor(hdc, bc);
        Windows.RestoreDC(hdc, dc);

        Windows.DeleteObject(hb5);
        Windows.DeleteObject(hb4);
        Windows.DeleteObject(hb3);
        Windows.DeleteObject(hb2);
        Windows.DeleteObject(hb);

    end;
    end;
end;



//
// And Item was selected by keystroke, or by click
//
procedure TFrmDummyUnicodePopup.WMMenuChar(var Msg: TWMMenuChar);
var
    s : string;
    //m : TWMCommand;
    mitd : TMenuItemTagData;
    mi : TMenuItem;
begin
    if (msg.MenuFlag and MF_POPUP) <> 0 then begin
        UnitMisc.AppendLog('wwmenuchar');

        self.TabCommand := false;

        //
        // Wait until user releases keystrok
        // Translate key to ItemID
        // pass information WMCommand - as if it was clicked by a mouse
        //

        while KeyboardQuery.IsPressed(VkKeyScan(msg.user)) do begin
            Application.ProcessMessages;
            sleep(10);
        end;
        msg.Result := MakeLong(0, MNC_CLOSE);


        s := KeyToItemID.Values[lowercase(msg.User)];

        self.InClipTypeIcon := false;
        if (s <> '') then begin
            //
            // trigger the correct WMCommand event
            //



            Application.ProcessMessages;
            m.NotifyCode := 0;
            m.ItemID := StrToInt(s);
            m.Ctl := 0;

            self.delayCommand := true;
            tim2.Enabled := true;
            //self.WMCommand(m);
        end else begin
            if msg.User = ' ' then begin
                m.ItemID := self.LastSelectID;
                m.NotifyCode := 0;
                m.Ctl := 0;

                self.delayCommand := true;
                tim2.Enabled := true;
            end else if msg.user = chr(vk_tab) then begin
                m.NotifyCode := 0;
                m.Ctl := 0;
                m.ItemID := self.LastSelectID;
                self.TabCommand := true;
                
                self.delaycommand := true;
                tim2.enabled := true;

            end;
        end;
    end else begin
        inherited;
    end;

    self.delayCommand := false;
end;


procedure TFrmDummyUnicodePopup.WMCommand(var Msg: TWMCommand);
var
    itemID : integer;
    mi2 : TMenuItem;
    mitd : TMenuItemTagData;
begin
    // Given an ItemID - carry out the selected action
    UnitMisc.AppendLog('command');

    if (msg.NotifyCode = 0) then begin
        itemID := msg.ItemID;
        mi2 := self.ItemIDToMenuItem(itemID);
        if (mi2 <> nil) then begin


            mitd := TagDataList.get(mi2.Tag);

            // Some items expect the menu item passed back
            // Other items need this form's handle to set focus correctly

            case mitd.itemtype  of
            IT_MENU, IT_TEMP:
                begin
                    if (self.InClipTypeIcon) or (self.TabCommand) then begin
                        FrmMainPopup.SendText('', mitd.clip, true);
                    end else if (self.RightClicked) then begin
                        FrmMainPopup.SetItemMenuOverride(true);
                        mi2.OnClick(mi2);
                    end else if (self.InApplicationIcon) then begin
                        FrmEditItem.SetText(mitd.clip.GetAsText, nil);
                        FrmMainPopup.ShowPreviewEditForm;
                    end else begin
                        mi2.onclick(mi2);
                    end;
                end;
            IT_PERMANENT:
                begin
                if mitd.itemtype in [IT_PERMANENT]  then begin
                    if self.RightClicked  then FrmMainPopup.SetItemMenuOverride(true);
                end;

                   mi2.OnClick(mi2);
                end;
            IT_CURRENT:
                begin
                    FrmMainPopup.CurrentMenuItemClickEvent(self, self.RightClicked);
                end;
            IT_SELECTED:
                begin
                    if self.RightClicked then FrmMainPopup.SetItemMenuOverride(true);
                    mi2.OnClick(mi2);
                end;
            else
                begin
                    mi2.OnClick(mi2);
                end;
            end;
            FrmMainPopup.SetItemMenuOverride(false);


        end;
    end else begin
        {dunno if this is needed}
        inherited;
    end;

    self.TabCommand := false;
end;




{
//
// These 3 messages control the menu popup
// Item is selected, Idle state occurs, then the timer will fire the
// showtooltip event
}

procedure TFrmDummyUnicodePopup.WMEnterIdle(var Msg: TWMEnterIdle);
var pos : TPoint;

    charmsg : TWMMenuChar;
    function equal(p1, p2 : TPoint) : boolean;
    begin
        result := p1.x = p2.x;
        result := result and (p1.y = p2.y);
    end;
begin
    //UnitMisc.AppendLog('wmenteridle');

    // Major "Clever" Trick
    // the popup menu doesn't response to right-clicks and other
    // keyboard presses, but I can test them on an "enter idle" event;

    self.RightClicked := false;
    if KeyboardQuery.IsClicked(rightButton) then begin
        self.RightClicked := true;
    end;


    tim.Interval := Application.HintPause;

    self.LastCursorPos  := mouse.CursorPos;
    pos := self.LastCursorPos;

    if (LastClipItem = nil) and (lasthint = '') then begin
        EXIT;
    end;

    if PasteTooltip.Showing then begin
       self.DisplayTootlip;
    end else begin
        inc(pos.X, 30);
        pos.Y := self.LastMenuRect.Bottom;

        InClipTypeIcon := false;
        InApplicationIcon := false;
        
        if self.MouseInCliptypeIcon then begin
            //FrmMenu.Close;
            Tooltip.CloseTooltip;
            PasteToolTip.ShowTooltip(
                'Press TAB or Click here to paste as ' + self.LastClipItem.GetFormatName,
               pos, false );
            InClipTypeIcon := true;
        end else if self.MouseInApplicationIcon  then begin

            Tooltip.CloseTooltip;
            PasteTooltip.ShowTooltip('Click here to Edit', pos, false);
            InApplicationIcon := true;
        end else begin
            //FrmMenu.Close;
            PasteTooltip.CloseTooltip;
            //if not tooltip.Showing  then begin
                tim.enabled := false;
                tim.Enabled := true;
            //end
        end;
    end;


    self.LastIdlePos := self.LastCursorPos;
    timclose.interval := 0;
    timclose.Interval := 2000;
    //TimClose.Enabled := true;


    if KeyboardQuery.IsPressed(VK_APPS) then begin
        self.RightClicked := true;
        Paste.SendSPACE;

        {
        charmsg.MenuFlag :=  MF_POPUP;
        charmsg.User := chr(VK_SPACE);
        charmsg.Msg := WM_MENUCHAR;
        self.WMMenuChar(charmsg);}

        {Application.ProcessMessages;
        m.NotifyCode := 0;
        m.ItemID := LastSelectID;
        m.Ctl := 0;

        self.delayCommand := true;
        tim2.Enabled := true;

        Windows.PostMessage(self.Handle, WM_NULL, 0, 0);
        }

    end;

end;

procedure TFrmDummyUnicodePopup.timTimer(Sender: TObject);
begin
    tim.Enabled := false;
    if self.TimeFiring then begin
        UnitMisc.AppendLog('Timer Conflict');
        self.TimeFiring := false;
        EXIT;
    end;
    self.TimeFiring  := true;
    UnitMisc.AppendLog('TimerFired');


    if (self.LastClipItem = nil) and (self.LastHint = '') then begin
        self.TimeFiring := false;
        EXIT;
    end;


    {if Windows.PtInRect(self.LastMenuRect, mouse.CursorPos) then begin
        self.LastCursorPos := mouse.CursorPos;
    end;
    }
    //inc(self.LastCursorPos.X, 20);
    //inc(self.LastCursorPos.Y, 10);
    self.DisplayTootlip;

    self.LastHint := '';
    self.TimeFiring := false;
end;

function TFrmDummyUnicodePopup.ItemIDToMenuPosition(
  itemID: cardinal): integer;
var mi : TMenuItem;
begin
    result := -1;
    mi := self.ItemIDToMenuItem(itemID);
    if (mi <> nil) and (mi.Parent <> nil) then begin
        result := mi.Parent.IndexOf(mi);
    end;
end;



procedure TFrmDummyUnicodePopup.WMMenuSelect(var Msg: TWMMenuSelect);
var id : word;
    i  : integer;
    mi : TMenuItem;
    //r : Trect;
    mitd : TMenuItemTagData;
begin
    //UnitMisc.AppendLog('wmselect');

    tim.Enabled := false;
    self.LastHint := '';
    self.LastClipITem := nil;
    self.LastIsMenuItem := false;


    //
    // Find the MenuItem by the unique ID and then use it to
    // find the actual index of the menu item
    //
    // Use the mouse position, if it's within the MenuItem otherwise
    // place it directly next to the menu
    //
    id := TWMMENUSELECT(msg).IDItem;
    LastSelectID := id;
    mi := self.ItemIDToMenuItem(id);
    if (mi <> nil) and (mi.Parent <> nil) then begin
        i := mi.Parent.IndexOf(mi);
        Windows.GetMenuItemRect(0, msg.Menu, i, self.LastMenuRect);
    end;


    //
    // ignore Menu items with submenus - they don't have tooltips
    //
    if ((TWMMENUSELECT(msg).MenuFlag and MF_POPUP) <> 0) then begin
        Tooltip.CloseTooltip;
        PasteTooltip.CloseTooltip;
        EXIT;
    end;


    if mi = nil then begin
        Tooltip.CloseTooltip;
        PasteTooltip.CloseTooltip;
        EXIT;
    end;

    if mi <> self.LastMenuItem then begin
        // need to redisplay when a new item is selected
        Tooltip.CloseTooltip;
        PasteTooltip.CloseTooltip;
    end;
    self.LastMenuItem := mi;
    PasteTooltip.CloseTooltip;

    //
    // gather the clip, the tooltip hint to show, and detect Menu Items
    //
    LastHotkeyName := '';
    try
    mitd := TagDataList.Get(mi.tag);

    case mitd.itemtype  of
        IT_NONE: ;
        IT_MENU:
            begin
                self.LastClipItem := mitd.clip;
                self.LastIsMenuItem := true;
            end;
        IT_OTHER:
            self.LastClipItem := mitd.clip;
        IT_PERMANENT:
            begin
                mitd := TagDataList.Get(mi.tag);

                self.LastHint := mitd.text;
                if (mitd.PermanentGroupID  <> - 1) then begin
                    FrmPermanent.PermFolderPush;
                    FrmPermanent.SetPermanentPath(FrmPermanent.PermFoldersGetItem(mitd.PermanentGroupID));
                    self.LastHint := mitd.text;

                    // clean up leaks
                    // GetComplexItem generates a new ClipItem every call
                    if FrmPermanent.IsComplexItem(self.LastHint) then begin
                        if self.LastClipItem <> nil then begin
                            myfree(self.lastclipitem);
                        end;
                        self.LastClipItem := FrmPermanent.GetComplexItem(mitd.text) ;
                    end;

                    LastHotkeyName := FrmPermanent.GetItemHotkeyName(mitd.itemindex);
                    FrmPermanent.PermFolderPop;
                end else begin
                    LastHotkeyName := FrmPermanent.GetItemHotkeyName(mitd.itemindex);
                end;
            end;

        IT_PERMANENT_GROUP: ;
        IT_POPUP: ;
        IT_TEMP:
            begin
                self.LastClipItem := mitd.clip;
                if (self.LastClipItem = nil) then begin
                    self.LastHint := mitd.text;
                end;
            end;

        IT_SWITCH: ;
        IT_SYS_SWITCH: ;
        IT_SYS_METHOD: ;
        IT_CURRENT:
            self.LastClipItem := self.CurrentClipboard;
        IT_LAST:
            self.LastHint := FrmMainPopup.GetLastStringSelected;
        IT_PROGRAM: ;
        IT_SELECTED:
            self.LastHint := FrmMainPopup.GetSelectedText;
    end;
    tim.Enabled := true;
    finally

    end;
end;


function TFrmDummyUnicodePopup.ItemIDToMenuItem(
  itemID: cardinal): TMenuItem;
var s : string;
    i : integer;
begin
    result := nil;

    // too spammy for normal use
   // UnitMisc.AppendLog('ItemIDToMenu: ' + IntToStr(ItemID));


    // either ItemID is a WM_COMMAND (an index to an object)
    // or it's an address of a menu item
    // - find the correct one and return it or NIL

    if (itemID < Cardinal(SubObjectList.Count)) then begin
        result := TMenuItem(SubObjectList.Items[itemID]);
    end;

    if result = nil then begin
        s := SubMenuItemList.values[IntToStr(itemID)];
        if (s <> '') then begin
            i := StrToInt(s);
            if i < SubObjectList.Count then begin
                result := TMenuItem( SubObjectList.items[i] );
            end;
        end;
    end;
end;

procedure TFrmDummyUnicodePopup.tim2Timer(Sender: TObject);
begin
    if (not self.delayCommand) then begin
        tim2.Enabled := false;
        self.WMCommand(m);
    end;
end;

procedure TFrmDummyUnicodePopup.TimCloseTimer(Sender: TObject);
begin
    timclose.Enabled := false;
    UnitMisc.AppendLog('TimClose');
end;


function TFrmDummyUnicodePopup.MouseInApplicationIcon: boolean;
begin
    result :=
        (Windows.PtInRect(self.LastMenuRect, mouse.CursorPos)) and
        (mouse.CursorPos.x < self.LastMenuRect.left + (ICON_SPACER-2)) and
        (self.LastClipItem <> nil) and self.LastIsMenuItem;
end;

function TFrmDummyUnicodePopup.MouseInCliptypeIcon: boolean;
begin
    result :=
        (Windows.PtInRect(self.LastMenuRect, mouse.CursorPos)) and
        (mouse.CursorPos.x > self.LastMenuRect.right - 15) and
        (self.LastClipItem <> nil) and
        not (self.LastClipItem.GetFormat in [CF_UNICODETEXT, CF_TEXT]);
end;

end.
